@@ -10,6 +10,6 @@ class Perf(object):
         self.time = time.time()
         self.logger.debug("ENTER %s %s", self.name, self.time)
 
-    def __exit__(self, exc_type, exc_value, traceback):
+    def __exit__(self, exc_type=None, exc_value=None, traceback=None):
         now = time.time()
         self.logger.debug("EXIT %s %s %s", self.name, now, now - self.time)
